#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <stdio.h>  
#include <stdlib.h> 
#include <vector>   
#include <map> 

#if defined(__MINGW32_VERSION) || defined(_MSC_VER) || defined(__BCPLUSPLUS__)
#include <process.h>
#define GNUPLOT_PATH string("\"C:\\Program Files\\gnuplot\\bin\\wgnuplot.exe\" ")
#else
#include <unistd.h>
#define GNUPLOT_PATH string("gnuplot ")
#endif


using namespace std;


//=========================================================
// Vector class
//=========================================================
class Vector : public vector<double>{
public:
  // coonstructor
  Vector( ) {}
  Vector( unsigned int n ) : vector<double> ( n ) {}
  Vector( const Vector& v ) : vector<double> ( v ) {}

  // insert new entry
  void pushBack( double x ){
    this->push_back( x );
  }
        
  // set all entries to r
  void setAll( double r ) {
  for ( unsigned int i = 0; i < size(); i++ )
      ( *this )[i] = r;
  }

  // set all entries to zero
  void setZero() {
    setAll( 0. );
  }

  // get squared l^2 norm
  double normSqr() const {
    double r = 0.0;
    for ( unsigned int i = 0; i < size(); i++ )
      r += ( *this )[i] * ( *this )[i];
    return r;
  }
	
  // get l^2 norm
  double norm() const {
    return std::sqrt( normSqr() );
  }

  // this += d * v, where d is a scalar and v is a vector (of same size as this!)
  void addMultiple  ( const Vector& v, const double d ){
    if( size() != v.size() )
      throw "Vector::addMultiple! wrong size!";
  
    for ( unsigned int i = 0; i < size(); i++ )
      (*this)[i] += d * v[i];
  }

  // multiplies every entry by r
  void operator*=( const double r ) {
    for ( unsigned int i = 0; i < size(); i++ )
      (*this)[i] *= r;
  }

  // divied every entry by r
  void operator/=( const double r ) {
    if( std::abs( r ) < 1e-15 )
      throw "Vector::operator/=! division by zero!";
    for ( unsigned int i = 0; i < size(); i++ )
      (*this)[i] /= r;
  }

  // adds another vector v (of same size!)
  void operator+=( const Vector& v ) {
    if( size() != v.size() )
      throw "Vector::operator+=! wrong size!";
    for ( unsigned int i = 0; i < size(); i++ )
      (*this)[i] += v[i];
  }
	
  // subtracts another vector v (of same size)
  void operator-=( const Vector& v ) {
    if( size() != v.size() )
      throw "Vector::operator-=! wrong size!";
    for ( unsigned int i = 0; i < size(); i++ )
      (*this)[i] -= v[i];
  }
	
  // returns dot product with vector v (of same size)
  double operator*( const Vector& v ) const {
    if( size() != v.size() )
      throw "Vector::operator*! wrong size!";
    double r = 0.0;
    for ( unsigned int i = 0; i < size(); i++ )
      r += v[i] * ( *this )[i];
    return r;
  }

  // write to console
  void print() const {
    for ( unsigned int j = 0; j < size(); j++ )
      cerr << (*this)[j] <<  " ";
    cerr << endl << endl;
  }

};

//=========================================================
// Sparse matrix class
//=========================================================
class SparseMatrix : protected std::vector< std::map<int, double> >{
  
  typedef std::map<int, double> SparseRow;
  unsigned int _m, _n;
    
public:
  // constructor
  SparseMatrix( unsigned int m, unsigned int n ) : vector<SparseRow>( m ), _m( m ), _n( n ) {}

  // get num of rows
  unsigned int getRows() const {
    return _m;
  }
	
  // get num of cols
  unsigned int getCols() const{ 
    return _n;
  }

  // read access
  double operator()( unsigned int i, unsigned int j ) const {
    checkRange( i, j, "readOperator" );
    SparseRow::const_iterator it = (*this)[i].find(j);
    if ( it != (*this)[i].end() ) return it->second;
    else return 0.0;
  }
	
  // write accesss
  double& operator()( unsigned int i, unsigned int j ) {
    checkRange( i, j, "writeOperator" );
    return (*this)[i][j];
  }
	
  // x = lambda * A * b, where A is this matrix, lambda a scalar and b a vector
  void apply( const Vector& x, Vector& b, double lambda = 1.0 ) const {
    if( b.size() != _m )
      throw "SparseMatrix::apply! dest vector has wrong size!";
    b.setZero();
    applyAdd(x, b, lambda);
  }
        
  // x += lambda * A * b, where A is this matrix, lambda a scalar and b a vector
  void applyAdd( const Vector& x, Vector& b, double lambda = 1.0 ) const {
    if( x.size() != _n )
      throw "SparseMatrix::applyAdd! arg vector has wrong size!";
    if( b.size() != _m )
      throw "SparseMatrix::applyAdd! dest vector has wrong size!";

    SparseRow::const_iterator iter;
    for ( unsigned int i = 0; i < _m; i++ )
      for ( iter = (*this)[i].begin(); iter != (*this)[i].end(); ++iter ) 
        b[i] += iter->second * x[iter->first] * lambda;
  }
	
  // x = lambda * A^T * b, where A is this matrix, lambda a scalar and b a vector
  void applyTransposed( const Vector& x, Vector& b, double lambda = 1.0 ) const {
    if( b.size() != _m )
      throw "SparseMatrix::applyTransposed! dest vector has wrong size!";
    b.setZero();
    applyAddTransposed(x, b, lambda);
  }
	
  // x = lambda * A^T * b, where A is this matrix, lambda a scalar and b a vector
  void applyAddTransposed( const Vector& x, Vector& b, double lambda = 1.0 ) const {
    if( x.size() != _m )
      throw "SparseMatrix::applyAddTransposed! arg vector has wrong size!";
    if( b.size() != _n )
      throw "SparseMatrix::applyAddTransposed! dest vector has wrong size!";

    SparseRow::const_iterator iter;
    for ( unsigned int j = 0; j < _n; j++ ){
      for ( unsigned int i = 0; i < _m; i++ ){
        SparseRow::const_iterator it = (*this)[i].find(j);
        if ( it != (*this)[i].end() )
          b[j] += it->second * x[i] * lambda;
      }
    }
  }
	
  // writes matrix to console
  void print() const {
    for ( unsigned int i = 0; i < _m; i++ ){
      for ( unsigned int j = 0; j < _n; j++ )
        cerr << get(i,j) <<  " ";
      cerr << endl;	      
    }
    cerr << endl;
  }
	
  // get min value of i-th row
  double getMinRowValue( int i ) const {
    if( !(i < _m) )
      throw "SparseMatrix::getMinRowValue! idx out of bounds!";

    SparseRow::const_iterator iter;
    double min = 0.;
    for ( iter = (*this)[i].begin(); iter != (*this)[i].end(); ++iter ) 
      if( iter->second < min )
        min = iter->second;
    return min;
  }

  // set entire matrix to zero
  void setZero() {
    for ( unsigned int i = 0; i < _m; i++ )
      (*this)[i].clear();
  }
	
  // set i-th row to zero
  void setRowZero( int i ) {
    if( !(i < _m) )
      throw "SparseMatrix::setRowZero! idx out of bounds!";
    (*this)[i].clear();
  }
	
  // set (i,j) entry
  void set( unsigned int i, unsigned int j, double entry ) {
    checkRange( i, j, "set" );
    if( std::abs( entry ) > 1e-15 )
      (*this)[i][j] = entry;
    else
     (*this)[i].erase(j);
  }
        
protected:
    // add to (i,j) entry
  void add( unsigned int i, unsigned int j, double entry ) {
	  checkRange( i, j, "add" );
	  if( std::abs( entry ) > 1e-15 )
	    (*this)[i][j] += entry;
  }
	
  // get (i,j) entry
  double get( unsigned int i, unsigned int j ) const {
	  checkRange( i, j, "get" );
	  SparseRow::const_iterator it = (*this)[i].find(j);
          if ( it != (*this)[i].end() ) return it->second;
          else return 0.0;
  }
	
  // check dimensions
  void checkRange(  unsigned int i, unsigned int j, string error ) const {
    if( !(i < _m) )
      throw "SparseMatrix::checkRange! row index too large! Error in " + error + "!";
    if( !(j < _n) )
      throw "SparseMatrix::checkRange! col index too large! Error in " + error + "!";
  }
  
};


//=========================================================
// CG solver class
//=========================================================
class CGSolver{

  double _tol;
  int    _maxiter;
	
public:      
  CGSolver() : _tol( 1e-15), _maxiter( 1000 ) {}
  CGSolver( double tol, int maxIter ) : _tol( tol ), _maxiter( maxIter ) {}
    
  // solve Ax = b for x if A is symmetric (default setting)
  // solve A^TA x = A^T b if A is not symmetric
  void solve( const SparseMatrix& A, const Vector & b, Vector & x, bool MatrixIsSymmetric = true ) const {         
      
    // residuum: r_0 = b - A * x (A is symmetric)
    Vector r( b );     
    A.applyAdd( x, r, -1.0 );   
     
    // residuum: r_0 = A^T (b - A x) (if A is NOT symmetric)
    if( !MatrixIsSymmetric ){
	Vector tmp( r );
	A.applyTransposed( tmp, r );
    }
      
    // descent direction d_0 = r_0
    Vector d( r );     

    double resNormSqr = r.normSqr();
    double error = std::sqrt( resNormSqr ); 
    int iter = 0;

    while ( ( iter < _maxiter) && ( error > _tol ) ) {
	// z = A * d_k (or z = A^T A * d_k, if A is NOT symmetric)
	Vector z( d.size() );   
        A.apply( d, z );        
	if( !MatrixIsSymmetric ){
	  Vector tmp( z );
	  A.applyTransposed( tmp, z );
        }
        
	// alpha = Norm(r_k)^2 / (d_k^T * A * d_k)  
        double alpha = resNormSqr / ( d * z );   
	// x_{k+1} = x_k - alpha * d_k
        x.addMultiple( d,  alpha );       
	// r_{k+1} = r_k + alpha * A * d_k
        r.addMultiple( z, -alpha );        

        double resNormSqr_old = resNormSqr;
        resNormSqr = r.normSqr();

        // update descent direction
        double beta = resNormSqr / resNormSqr_old;
        // d_{k+1} = beta * d_k + r_{k+1}
        d *= beta;
        d += r;

        error = std::sqrt( resNormSqr );
        iter++;
     }
     
     if ( error > _tol ) 
       cerr << endl << "!!! Could not solve to required tolerance ( " << _tol << " ) within " << _maxiter << " iterations !!!" << endl << endl;
     else
       cerr << endl << "Solved to required tolerance ( " << _tol << " ) within " << iter << " iterations !!!" << endl << endl;
    }

    void setTol( double tol ) {
	_tol = tol;
    }
      
    void setMaxIter( int maxiter ) {
	_maxiter = maxiter;
    }

};


//=========================================================
// Triangular mesh
//=========================================================
class TriangMesh{
  
int _numOfNodes, _numOfFaces;  
std::vector< Vector > _coords;
std::vector< Vector > _triang;
mutable std::vector<bool> _bdryIdx;
  
public:
  // construct empty triangular mesh
  TriangMesh( ) : _numOfNodes(0), _coords(2), _triang(3) {}
  
  // construct triangular mesh from file
  TriangMesh( string objFilename ) : _numOfNodes(0), _coords(2), _triang(3) {
    loadFromOBJ( objFilename );
  }
  
  // get number of nodes
  int getNumberOfNodes() const {
    return _numOfNodes;
  }
  
  // get number of faces
  int getNumberOfFaces() const {
    return _numOfFaces;
  }
  
  // get global node idx
  int getGlobalIdx( int triangIdx, int localIdx ) const {
    if( !(triangIdx < _numOfFaces) )
      throw "TriangMesh::getGlobalIdx(): global Idx out of bounds!";
    if( localIdx > 2 )
      throw "TriangMesh::getGlobalIdx(): local Idx out of bounds!";
    return _triang[localIdx][triangIdx];
  }
  
  // get x- or y-coordinate of i-th node (coord = 0 or coord = 1, resp.)
  double getVertexCoord( int i, int coord ) const {
    if( !(i < _numOfNodes) )
      throw "TriangMesh::getVertexCoord(): global Idx out of bounds!";
    if( coord > 1 )
      throw "TriangMesh::getVertexCoord(): coord out of bounds!";
    return _coords[coord][i];
  }
  
  //
  void getVertexCoord( int nodeIdx, Vector& coords ) const {
    if( !(nodeIdx < _numOfNodes) )
      throw "TriangMesh::getVertexCoord(): global Idx out of bounds!";
    coords.clear();
    for( int i = 0; i < 2; i++ )
      coords.pushBack( _coords[i][nodeIdx] );
  }
  
  // get area of i-th triangle
  double getArea( int i ) const {
    if( !(i < _numOfFaces) )
      throw "TriangMesh::getWeightedNormal(): global Idx out of bounds!";
    Vector wn;
    getWeightedNormal( i, wn );
    return wn.norm() / 2.;
  }

  // returns "true" if i-th vertex is boundary node ("false" otherwise)
  bool isBoundaryNode( int i ) const {
    if( !(i < _numOfNodes) )
      throw "TriangMesh::isBoundaryNode(): global Idx out of bounds!";
    if( _bdryIdx.size() != _numOfNodes )
      setBoundaryNodes();
    return _bdryIdx[i];
  }
  
  // load from an obj file
  void loadFromOBJ ( string filename ) {
    ifstream in ( filename.c_str() );
    
    _numOfNodes = 0;
    _numOfFaces = 0;
    _coords.clear();
    _triang.clear();
  
    do {
      string temp;
      char line[256];
      in >> temp;
      // skip line of comments
      if ( temp.compare ( "#" ) == 0 ){        
        in.getline ( line, 256 );
        continue;
      }
      // otherwise expect data
      if ( temp.compare ( "v" ) == 0 ) {
	_numOfNodes++;
        for ( int i = 0; i < 2; ++i ){	  
	  double val;
          in >> val;
	  _coords[i].pushBack( val );
	}
	in.getline ( line, 256 );
        continue;
	
      }
      else if ( temp.compare ( "f" ) == 0 ) {
	_numOfFaces++;
        for ( int i = 0; i < 3; ++i ) {
	  int idx;
          in >> idx;
          _triang[i].push_back( idx - 1 );
        }	
      }  
      else if ( temp.length() > 0 )
        throw "TriangMesh::loadFromOBJ(): Error parsing the input file.";
    }
    while ( in.eof() == false );

  }
  
protected:    
    
  // get weighted face normal
  double getWeightedNormal( int triangIdx, Vector& wn ) const {
    if( !(triangIdx < _numOfFaces) )
      throw "TriangMesh::getWeightedNormal(): global Idx out of bounds!";
    
    // get edges
    Vector e0(3), e1(3);
    int nodeIdx0 = getGlobalIdx(triangIdx, 0);
    int nodeIdx1 = getGlobalIdx(triangIdx, 1);
    int nodeIdx2 = getGlobalIdx(triangIdx, 2);
    
    for( int j = 0; j < 2; j++ ){
      e0[j] = getVertexCoord( nodeIdx1, j ) - getVertexCoord( nodeIdx0, j );
      e1[j] = getVertexCoord( nodeIdx2, j ) - getVertexCoord( nodeIdx0, j );
    }
    
    // wn = e0 x e1
    wn.clear();
    for( int j = 0; j < 3; j++ )
      wn.pushBack( e0[ (j+1)%3 ] * e1[ (j+2)%3 ] - e0[ (j+2)%3 ] * e1[ (j+1)%3 ] );
  }
  
  // compute boundary nodes
  void setBoundaryNodes() const {
    SparseMatrix adjMat( _numOfNodes, _numOfNodes );
     _bdryIdx.resize( _numOfNodes );
     
    // build adjacency matrix: -1 means boundary edge
    for( int i = 0; i < _numOfFaces; i++ ){
      for( int j = 0; j < 3; j++ ){
	int min = std::min( getGlobalIdx(i, j), getGlobalIdx(i, (j+1)%3) );
	int max = std::max( getGlobalIdx(i, j), getGlobalIdx(i, (j+1)%3) );
	if( adjMat( min, max ) == 0 )
	  adjMat( min, max ) =  adjMat( max, min ) = -1.;
	else
	  adjMat( min, max ) =  adjMat( max, min ) = 1.;
      }
    }  
    
    // boundary node if at least one adjacent edge is boundary edge
    for( int i = 0; i < _numOfNodes; i++ ){
      if( adjMat.getMinRowValue(i) < 0 )
	_bdryIdx[i] = true;
    }    
  }
};

// plots vector on grid
void plotPNG( const TriangMesh& triangMesh, const Vector & sol, string filename ) {
    
    if ( sol.size() < static_cast<size_t>( triangMesh.getNumberOfNodes() ) )  {
        cerr << "Solution Vector does not have enough entries!" << endl;
        return;
    }

    fstream file;
    stringstream ss; ss << getpid();
    string tmpfile = ".gp_" + ss.str() + ".tmp";
    file.open( tmpfile.c_str(), fstream::out );
    if ( file.is_open() ){
      file << "set terminal png large size 800,800 enhanced" << endl;
      file << "set output \"" << filename << "\"" << endl;
      file << "set pm3d interpolate 10,10 implicit map" << endl;
      file << "set palette model HSV" << endl;
      file << "set palette defined (0 0.67 1 0.7, 0.33 0.5 1 0.8, 0.67 0.166 1.0 0.9, 1 0 1 1)" << endl;
      file << "set colorbox user size .02,.6" << endl;
      file << "unset border" << endl;
      file << "set xtics nomirror" << endl;
      file << "set ytics nomirror" << endl;
      file << "splot '-' notitle with dots" << endl;

      for ( int i = 0; i < triangMesh.getNumberOfFaces(); i++ ){
	file << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 0), 0 ) << " " << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 0), 1 ) << " " << sol[triangMesh.getGlobalIdx(i, 0)] << endl;
        file << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 1), 0 ) << " " << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 1), 1 ) << " " << sol[triangMesh.getGlobalIdx(i, 1)] << endl << endl;
        file << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 0), 0 ) << " " << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 0), 1 ) << " " << sol[triangMesh.getGlobalIdx(i, 0)] << endl;
        file << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 2), 0 ) << " " << triangMesh.getVertexCoord( triangMesh.getGlobalIdx(i, 2), 1 ) << " " << sol[triangMesh.getGlobalIdx(i, 2)] << endl << endl << endl;
      }
      file << "e" << endl;

      file.close();
      system( (GNUPLOT_PATH + tmpfile).c_str() );
      remove( tmpfile.c_str() );
  }
  else{
    cerr << "Error opening file " << filename << endl;
  }
}